/**Panel4P
 * Minimal Interface/Controller
 * 
 *Copyright (c) 2011 Ioannis (Yiannis) Chatzikonstantinou, All rights reserved.
 *http://www.volatileprototypes.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */

package volatileprototypes.Panel4P;

import processing.core.*;
import java.awt.*;
import javax.swing.*;
import java.util.ArrayList;
import java.lang.reflect.*;
import java.awt.event.*;
import java.util.*;

public class Panel4P extends JPanel implements MouseListener, MouseMotionListener, MouseWheelListener, PanelStyles {
    protected AutoFrame frame;
    protected PApplet app;
    protected ArrayList<Component> components = new ArrayList<Component>();
    protected Component activeComponent = null;
    
    private boolean updateField = true;
	
    private float scrollSpeed = 0;
	private float scrollAcc = 0;
	private java.util.Timer scrollTimer;
	
    private int maxHeight = 800;
    protected int width = 180;
    protected int height = 480;
	protected float yOffset = -2;
    protected int totalElementsHeight = 0;
	
	protected Graphics2D g2;
    
    protected int mouseX = 0, mouseY = 0, pmouseX = 0, pmouseY = 0;
    
    public Panel4P (PApplet appin) {
        app=appin;
		
        frame = new AutoFrame(this);
		frame.getContentPane().add(this);
		frame.setSize(width,height);
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
        app.registerPre(this);
        System.out.println("Panel4P version 0.8.3");
	    
		frame.setVisible(true);
    }
    
    //
    // Before the parent app's draw() method, trigger any linked methods and set
    // any linked fields.
    //
    
    public void pre() {
      if (activeComponent!=null) {
        // If specified, try calling the component's trigger method.
        setField(activeComponent);
        // If specified, try setting the component's field to the new value
        callMethod(activeComponent);
        activeComponent=null;
      }
    }
  
    protected void callMethod(Component c) {
        if (c.trigger!=null) {
            try {
                Class[] t=c.trigger.getParameterTypes();
                if (t.length==0) {
                    c.trigger.invoke(app);
                } else if (t[0]==boolean.class) {
                    boolean b=(c.value != 0.0f);
                    c.trigger.invoke(app, b);
                } else if (t[0]==int.class) {
                    c.trigger.invoke(app, (int)c.value);
                } else if (t[0]==float.class) {
                    c.trigger.invoke(app, c.value);
                } else if (t[0]==double.class) {
                    c.trigger.invoke(app, (double)c.value);
                }
            } catch (Exception e) {
                System.err.println("Error while calling method \'" + c.name + "\'. Disabling.");
                e.printStackTrace();
                c.trigger=null;
            }
        }
    }
  
    protected void setField(Component c) {
        if (c.field!=null) {
            try {
                Class t=c.field.getType();
                if (t==boolean.class) {
                    boolean b=(c.value != 0.0f);
                    c.field.setBoolean(app, b);
                } else if (t==int.class) {
                    c.field.setInt(app, ((int)c.value));
                } else if (t==float.class) {
                    c.field.setFloat(app, c.value);
                }
            } catch (Exception e) {
                System.err.println("Error while seting field \'" + c.name + "\'. Disabling.");
                e.printStackTrace();
                c.field=null;
            }
        }
    }
    
    // Rearranges the elements, calculating again their positions
    // and updating the total height.
    private final void rearrange() {
        int y=0;
        for (Component c : components) {
            y=c.updateY(y);
        }
        totalElementsHeight=y+3;
        repaint();
    }
	
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		this.g2 = (Graphics2D) g;
		for (Component c : components) {
            c.rdraw();
        }
	}
	
	//
	// Resize
	//
	
	public final void autoSize() {
        height = Math.min(maxHeight, totalElementsHeight);
        frame.setSize(width,height);
        rearrange();
    }
	
	public void setMaxHeight(int nMaxHeight) {
	  maxHeight = Math.max(150,nMaxHeight);
	}
	
	protected void doResize() {
	    height = Math.min(frame.getSize().height, Math.min(maxHeight, totalElementsHeight));
		frame.setSize(width,height);
		rearrange();
	}
    
    //
    // Public classes to correctly add & register components
    //
  
    // Single Button
    public void addButton(String namein, String label) {
        Component b = new Button(this, namein, label);
        components.add(b);
        rearrange();
        try {
            b.setTrigger(app.getClass().getMethod(namein));
        } catch (Exception e) {}
        try {
            b.setField(app.getClass().getField(namein));
        } catch (Exception e) {}
    }
  
    // Single toggle button
    public void addButton(String namein, String label, float sticky) {
        Component b = new Button(this, namein, label, sticky);
        components.add(b);
        rearrange();
        try {
            b.setTrigger(app.getClass().getMethod(namein, new Class[] { boolean.class }));
        } catch (Exception e) {}
        try {
            b.setField(app.getClass().getField(namein));
        } catch (Exception e) {}
    }
  
    // Button Group
    public void addButtonGroup(String namein, String[] labels, float value) {
        if (labels.length>1) {
            Component b = new ButtonGroup(this, namein, labels, value);
            components.add(b);
            rearrange();
            try {
                b.setTrigger(app.getClass().getMethod(namein, new Class[] { int.class }));
            } catch (Exception e) {}
            try {
              b.setField(app.getClass().getField(namein));
            } catch (Exception e) {}
        } else if (labels.length==1) {
            addButton(namein, labels[0], 1);
        }
    }
  
    // Slider
    public void addSlider(String namein, String label, float min, float max, float value) {
        Component s = new Slider(this, namein, label, min, max, value);
        components.add(s);
        rearrange();
        try {
            s.setTrigger(app.getClass().getMethod(namein, new Class[] { float.class }));
        } catch (Exception e) {}
        try {
            s.setField(app.getClass().getField(namein));
        } catch (Exception e) {}
    }
  
    // Slider with custom roundoff at value display
    public void addSlider(String namein, String label, float min, float max, float value, int ndigits) {
        Component s = new Slider(this, namein, label, min, max, value, ndigits);
        components.add(s);
        rearrange();
        try {
            s.setTrigger(app.getClass().getMethod(namein, new Class[] { float.class }));
        } catch (Exception e) {}
        try {
            s.setField(app.getClass().getField(namein));
        } catch (Exception e) {}
    }

    // Label
    public void addLabel(String namein, String label) {
        LabelBox l = new LabelBox(this, namein, label);
        components.add(l);
        rearrange();
    }
  
    //Remove by name
    public void remove(String name) {
        ArrayList<Component> r = new ArrayList<Component>();
        for (Component c : components) {
            if (c.name.equals(name)) {
                r.add(c);
            }
        }
        components.removeAll(r);
        rearrange();
    }
    
    //
	// Responses to interface events
	//
	
	public void mousePressed(MouseEvent e) {
        updateMousePos(e);
        for (Component c : components) {
            c.checkForPressed();
        }
		repaint();
	}
  
	public void mouseReleased(MouseEvent e) {
        updateMousePos(e);
        for (Component c : components) {
            c.checkForReleased();
        }
		repaint();
	}
  
	public void mouseDragged(MouseEvent e) {
        updateMousePos(e);
        for (Component c : components) {
            c.checkForDrag();
        }
        if (activeComponent==null) {
            if (height<totalElementsHeight+20) {
                moveOffset(mouseY-pmouseY);
            }
        }
		repaint();
	}
    
    public void mouseMoved(MouseEvent e) {
        updateMousePos(e);
    }
	
	public void mouseWheelMoved(MouseWheelEvent e) {
	  scrollAcc += (float)e.getWheelRotation() / (float)e.getScrollAmount();
	  if (scrollTimer == null) {
		scrollTimer = new java.util.Timer("ScrollTimer");
		scrollTimer.schedule(new doScroll(), 5L, 35L);
	  }
	}
	
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	
    private void updateMousePos(MouseEvent e) {
		pmouseX = mouseX;
		pmouseY = mouseY;
        mouseX = e.getX();
        mouseY = e.getY();
    }
	
	private class doScroll extends TimerTask {
	  public void run() {
		scrollSpeed = scrollSpeed * 0.76f + scrollAcc;
		moveOffset(-scrollSpeed);
		scrollAcc *= 0.24;
		if (scrollSpeed < 0.02 && scrollSpeed > -0.02) {
		  scrollSpeed = 0;
		  scrollAcc = 0;
		  scrollTimer.cancel();
		  scrollTimer = null;
		}
		repaint();
	  }
	}
  
	//
	// Value getters and setters
	//
    
	public void moveOffset(float Y) {
	  yOffset = Math.max(Math.min(-2 + Y*0.2f, yOffset + Y),Math.min(-2,height-(totalElementsHeight+20)) + Y*0.2f);
	}
	
	public void setValue(String name, float val) {
        for (Component c : components) {
            if (name.equals(c.name)) {
                c.value=val;
                if (updateField) setField(c);
                repaint();
            }
        }
	}
  
	public float getValue(String name) {
        for (Component c : components) {
            if (name.equals(c.name)) {
                return(c.value);
            }
        }
        return(-1);
	}
	
	//
	// Update all linked variables
	//
	public void updateVars() {
        for (Component c : components) {
            setField(c);
        }
	}
  }